local codec     = require "kong.openid-connect.codec"
local jwa       = require "kong.openid-connect.jwa"


local base64url = codec.base64url
local tostring  = tostring
local concat    = table.concat
local upper     = string.upper
local json      = codec.json
local type      = type
local find      = string.find
local sub       = string.sub


local function split(jwt)
  local t = {}
  local i = 1
  local b = 1
  local e = find(jwt, ".", b, true)
  while e do
    t[i] = sub(jwt, b, e - 1)
    i = i + 1
    b = e + 1
    e = find(jwt, ".", b, true)
  end
  t[i] = sub(jwt, b)
  return t, i
end


local jws = {}


function jws.encode(token)
  if type(token) ~= "table" then
    return nil, "token needs to be a table"
  end

  local jwk = token.jwk
  if not jwk then
    jwk = {
      alg = "RS256",
    }
  end

  jwk.alg = jwk.alg or "RS256"

  local header = token[1] or token.header
  if not header then
    header = {
      typ = "JWT",
      alg = jwk.alg,
      kid = jwk.kid,
    }
  end

  local header_encoded, payload_encoded, err

  if type(header) == "table" then
    header_encoded, err = json.encode(header)
    if not header_encoded then
      return nil, "unable to json encode jws header (" .. err .. ")"
    end
  else
    header_encoded = tostring(header)
  end

  header_encoded, err = base64url.encode(header_encoded)
  if not header_encoded then
    return nil, "unable to base64 encode jws header (" .. err .. ")"
  end

  local payload = token[2] or token.payload
  if not payload then
    payload = {}
  end

  if type(payload) == "table" then
    payload_encoded, err = json.encode(payload)
    if not payload_encoded then
      return nil, "unable to json encode jws payload (" .. err .. ")"
    end
  else
    payload_encoded = tostring(payload)
  end

  payload_encoded, err = base64url.encode(payload_encoded)
  if not payload_encoded then
    return nil, "unable to base64 encode jws payload (" .. err .. ")"
  end

  local message = concat({ header_encoded, payload_encoded }, ".")
  local digest
  digest, err = jwa.sign(header.alg, jwk, message)
  if not digest then
    return nil, err
  end

  return concat({ message, digest }, ".")
end


function jws.decode(token, options, oic)
  options = options or {}
  oic     = oic     or { options = {} }

  local t = type(token)
  if t ~= "table" then
    if t ~= "string" then
      return nil,  "token needs to be a table or a string"
    end

    local z
    token, z = split(token)
    if z ~= 3 then
      return nil, "invalid jws token was specified"
    end
  end

  local header  = token[1] or token.header
  local payload = token[2] or token.payload
  local sig     = token[3] or token.signature

  local hdr, pld, ok, err

  hdr, err = base64url.decode(header)
  if not hdr then
    return nil, "unable to base64url decode jws header (" ..  err .. ")"
  end

  hdr, err = json.decode(hdr)
  if not hdr then
    return nil, "unable to json decode jws header (" ..  err .. ")"
  end

  local jwk

  local verify_signature = true
  if options.verify_signature ~= nil then
    verify_signature = not not options.verify_signature
  elseif oic.options.verify_signature ~= nil then
    verify_signature = not not oic.options.verify_signature
  end

  if verify_signature then
    if hdr.typ then
      local upper_typ = upper(hdr.typ)
      if upper_typ ~= "JWT" and upper_typ ~= "AT+JWT" then
        return nil, "only jwts are supported for the jws"
      end
    end

    local alg = hdr.alg
    if not alg then
      return nil, "jws algorithm was not specified"
    end

    local keys1 = options.keys or {}
    local keys2 =     oic.keys or {}

    local kid = hdr.kid
    if kid then
      jwk = keys1[kid]
      if not jwk then
        jwk = keys2[kid]
      end

      if not jwk and keys1.rediscover then
        keys1, err = keys1:rediscover()
        if err then
          return nil, err
        end

        if type(keys1) == "table" then
          jwk = keys1[kid]
        end
      end

      if not jwk and keys2.rediscover then
        keys2, err = keys2:rediscover()
        if err then
          return nil, err
        end

        if type(keys2) == "table" then
          jwk = keys2[kid]
        end
      end
    end

    if not jwk then
      if jwa.HMAC[alg] then
        jwk = keys1[alg] or keys1.default
        if not jwk then
          jwk = keys2[alg] or keys2.default
        end

      else
        if not jwk then
          jwk = keys1[alg]
          if not jwk then
            jwk = keys2[alg]
          end
        end
      end

      if not kid and not jwk and keys1.rediscover then
        keys1, err = keys1:rediscover()
        if err then
          return nil, err
        end

        if type(keys1) == "table" then
          jwk = keys1[alg]
        end
      end

      if not kid and not jwk and keys2.rediscover then
        keys2, err = keys2:rediscover()
        if err then
          return nil, err
        end

        if type(keys2) == "table" then
          jwk = keys2[alg]
        end
      end
    end

    if not jwk then
      return nil, "suitable jwk was not found (" .. (kid or alg) .. ")"
    end

    local inp = concat({ header, payload }, ".")

    ok, err = jwa.verify(alg, jwk, inp, sig)
    if not ok then
      return nil, err
    end
  end

  pld, err = base64url.decode(payload)
  if not pld then
    return nil, "unable to base64 url decode jws payload (" ..  err .. ")"
  end

  pld, err = json.decode(pld)
  if not pld then
    return nil, "unable to json decode jws payload (" ..  err .. ")"
  end

  return {
    jwk       = jwk,
    type      = "JWS",
    header    = hdr,
    payload   = pld,
    signature = sig,
  }
end


return jws
