local inspect = require "inspect"

local runloop_handler = require "kong.runloop.handler"
local sandbox = require "kong.tools.sandbox"

local workspaces = require "kong.workspaces"

local BasePlugin = require "kong.plugins.base_plugin"

local PLUGIN_NAME    = require("kong.plugins.exit-transformer").PLUGIN_NAME
local PLUGIN_VERSION = require("kong.plugins.exit-transformer").PLUGIN_VERSION


local function get_conf()
  -- Gets plugin configuration for the ctx, no matter the priority
  local workspace = workspaces.get_workspace()
  -- Not really needed, but a hack because get_workspace might return an
  -- empty {} to signal... something? AFAIK, this is fixed on 2.0 already,
  -- this solution makes it so it won't work on neither version of kong.
  local workspace_id = workspace and workspace.id

  local plugins_iterator = runloop_handler.get_plugins_iterator()

  for plugin, plugin_conf in plugins_iterator:iterate("access", ngx.ctx) do

    if plugin.name ~= PLUGIN_NAME then
      goto continue
    end

    -- it's very important that this filtering happens here and not once we
    -- already have a config. Since plugin confs applying globally on
    -- different workspaces would collide here and rely only on the first
    -- match
    if not workspace_id and not plugin_conf.handle_unknown then
      goto continue
    end

    if ngx.ctx.KONG_UNEXPECTED and not plugin_conf.handle_unexpected then
      goto continue
    end

    do
      return plugin_conf
    end

    ::continue::
  end

  return nil
end

local function_cache = setmetatable({}, { __mode = "k" })

local function no_op(err)
  return function(status, body, headers)
    kong.log.err(err)
    return status, body, headers
  end
end

local function get_functions(conf)
  if function_cache[conf] then return function_cache[conf] end

  local functions = {}

  local sandbox_opts = { env = { kong = kong } }

  for _, fn_str in ipairs(conf.functions) do
    local f, err = sandbox.validate_function(fn_str, sandbox_opts)
    if err then f = no_op(err) end

    table.insert(functions, f)
  end

  function_cache[conf] = functions

  return functions
end

local _M = BasePlugin:extend()

_M.PRIORITY = 9999
_M.VERSION = PLUGIN_VERSION

function _M:new()
  _M.super.new(self, PLUGIN_NAME)
end

function _M:init_worker()
  kong.response.register_hook("exit", self.exit, self)
end


function _M:access(conf)
  _M.super.access(self)
end


function _M:exit(status, body, headers)
  -- Do not transform admin requests
  if not ngx.ctx.is_proxy_request then
    return status, body, headers
  end

  -- Try to get plugin configuration for current context
  local conf = get_conf()
  if not conf then
    return status, body, headers
  end

  local functions = get_functions(conf)

  -- Reduce on status, body, headers through transform functions
  for _, fn in ipairs(functions) do
    status, body, headers = fn(status, body, headers)
  end

  return status, body, headers
end

return _M
